from unittest.mock import patch, MagicMock
from tuxbake.models import OEBuild


@patch.multiple(
    OEBuild,
    prepare=MagicMock(),
    do_cleanup=MagicMock(),
    do_build=MagicMock(return_value="build_called"),
    publish_artifacts=MagicMock(return_value="publish_artifacts_called"),
)
def test_build(oebuild_repo_init_object):
    from tuxbake.build import build

    oebuild_repo_init_object.local_manifest = None
    oebuild_object = oebuild_repo_init_object.as_dict()

    data = build(**oebuild_object)
    assert data == OEBuild(**oebuild_object)
    ret_val = data.do_build()
    assert ret_val == "build_called"
    ret_val = data.publish_artifacts()
    assert ret_val == "publish_artifacts_called"



import os
import json
import shutil
import subprocess
from typing import List
from pathlib import Path
from collections import OrderedDict
from configparser import ConfigParser as OrigConfigParser

from tuxbake import __version__
from tuxbake.exceptions import UnsupportedMetadata, UnsupportedMetadataType


def run_command(cmd, src_dir=None):
    try:
        out = subprocess.check_output(cmd, cwd=src_dir, text=True).strip()
        return out
    except Exception as e:
        print(e)


class ConfigParser(OrigConfigParser):
    def optionxform(self, opt):
        return str(opt)


class Metadata:
    basedir = "metadata"

    def __init__(self, oebuild):
        self.oebuild = oebuild
        self.runtime = self.oebuild._runtime
        self.src_dir = self.oebuild.src_dir
        self.build_dir = Path(f"{self.runtime.source_dir}/{self.oebuild.build_dir}")
        self.metadata = OrderedDict()
        self.files = self.meta_files()
        self.types = {}

    def collect(self):
        self.metadata["build"] = {
            "envsetup": self.oebuild.envsetup,
            "targets": self.oebuild.targets,
            "distro": self.oebuild.distro,
            "name": self.oebuild.name,
            "machine": self.oebuild.machine,
            "container": self.oebuild.container,
            "environment": self.oebuild.environment,
            "local_conf": self.oebuild.local_conf,
            "bblayers_conf": self.oebuild.bblayers_conf,
            "runtime": self.oebuild.runtime,
            "image": self.oebuild.image,
            "sstate_dir": self.oebuild.sstate_dir,
            "sstate_mirror": self.oebuild.sstate_mirror,
            "dl_dir": self.oebuild.dl_dir,
            "local_manifest": self.oebuild.local_manifest,
            "pinned_manifest": self.oebuild.pinned_manifest,
            "artifacts": self.oebuild.artifacts,
        }
        self.metadata["tuxbake"] = {"version": __version__}
        self.metadata["runtime"] = self.get_runtime_info()
        self.metadata["git"] = self.get_git_info()
        # collect data by running perl script
        self.metadata.update(self.generate_meta_dict())
        if self.metadata.get("hardware"):
            self.metadata["hardware"]["free_disk_space"] = self.free_disk_space()
        if hasattr(self.oebuild, "result"):
            self.metadata["results"] = {
                "status": "PASS" if self.oebuild.result == "pass" else "FAIL",
            }
        # save metadata.json file
        self.save_metadata()

    def get_runtime_info(self):
        if self.runtime.name in ["null", "docker-local", "podman-local"]:
            return {}

        return self.runtime.get_metadata()

    def save_metadata(self):
        with (Path(self.src_dir) / "metadata.json").open("w") as f:
            f.write(json.dumps(self.metadata, indent=4, sort_keys=True))
            f.write("\n")

    def meta_files(self) -> List[str]:
        # get all available '*.ini' files from metadata folder
        files = (Path(__file__).parent / self.basedir).glob("*.ini")
        return [str(f.name).replace(".ini", "") for f in files]

    def generate_meta_dict(self):
        # Generate metadata extractions pecification in JSON format to pass to script file
        # then process the script output and then return metadata dictionary
        metadata_input_data = {}
        for file in self.files:
            metadata_input_data[file] = {}
            conffile = Path(__file__).parent / self.basedir / f"{file}.ini"
            config = ConfigParser()
            config.read(conffile)
            # parse types field
            try:
                for k, t in config["types"].items():
                    if t not in ["int", "str", "linelist"]:
                        raise UnsupportedMetadataType(t)
                    self.types[k] = eval(t)
            except KeyError:
                pass
            # parse commands field
            try:
                commands = {
                    k: self.replace_placeholders(v)
                    for k, v in dict(config["commands"]).items()
                }
                metadata_input_data[file] = commands
            except KeyError as key:
                raise UnsupportedMetadata(f"{key} is required in '{file}.ini' file")

        metadata_input = self.build_dir / "metadata.in.json"
        metadata_input.write_text(json.dumps(metadata_input_data), encoding="utf-8")

        script_src = Path(__file__).parent / "metadata.pl"
        script = self.build_dir / "metadata.pl"
        shutil.copy(script_src, script)

        stdout = Path(self.build_dir / "extracted-metadata.json")
        with stdout.open("w") as f:
            self.runtime.run_cmd(
                ["perl", str(script), str(metadata_input)],
                echo=False,
                stdout=f,
                offline=False,
            )
        metadata = self.read_json(stdout.read_text())
        return metadata

    def replace_placeholders(self, key):
        return key.format(build_dir=self.build_dir)

    def read_json(self, metadata_json):
        # Loads the json, typecast its values into desired datatype and return result dict.
        if not metadata_json:
            return {}
        try:
            metadata = json.loads(metadata_json)
        except json.JSONDecodeError:
            return {"invalid_metadata": metadata_json}
        if not metadata:
            return {}

        result = {}
        for key in metadata:
            for k, v in metadata[key].items():
                if v:
                    v = v.strip()
                if v:
                    result.setdefault(key, {})
                    result[key][k] = self.cast(k, v)

        return result

    def cast(self, key, v):
        # typecast the command output with desired datatype
        t = self.types.get(key, str)
        return t(v)

    def free_disk_space(self):
        disk_usage = shutil.disk_usage(self.build_dir)
        return int(disk_usage.free / (2**20))

    def get_git_info(self):
        git_cmds = {
            "branch": "git rev-parse --abbrev-ref HEAD",
            "commit": "git rev-parse HEAD",
            "url": "git remote get-url origin",
        }
        ret = {}
        oebuild = self.oebuild
        if oebuild.sources.get("repo"):
            for key, cmd in git_cmds.items():
                ret[key] = run_command(cmd.split(), f"{self.src_dir}/.repo/repo")
        else:
            for git_object in oebuild.git_trees:
                temp = {}
                url = git_object.url.rstrip("/")
                dest = git_object.dest
                basename = os.path.splitext(os.path.basename(url))[0]
                if dest:
                    abs_dest = (
                        os.path.abspath(
                            os.path.join(self.src_dir, os.path.expanduser(dest))
                        )
                        + os.sep
                    )
                    dir_path = Path(os.path.join(abs_dest, basename))
                else:
                    dir_path = Path(os.path.join(self.src_dir, basename))

                for key, cmd in git_cmds.items():
                    temp[key] = run_command(cmd.split(), dir_path)
                ret[basename] = temp
        return ret



#!/usr/bin/python3
# -*- coding: utf-8 -*-

from tuxbake.models import OEBuild
from tuxbake.metadata import Metadata
from tuxmake.runtime import Terminated
import signal
import sys


def build(**kwargs):
    old_sigterm = signal.signal(signal.SIGTERM, Terminated.handle_signal)
    oebuild = OEBuild(**kwargs)
    try:
        oebuild.prepare()
        oebuild.do_build()
    except (KeyboardInterrupt, Terminated):
        print("tuxbake Interrupted")

    if oebuild.artifacts_dir:
        oebuild.publish_artifacts()
    # try to collect metadata
    try:
        metadata = Metadata(oebuild)
        metadata.collect()
    except Exception as ex:
        sys.stderr.write(f"{str(ex)}\n")

    oebuild.do_cleanup()
    signal.signal(signal.SIGTERM, old_sigterm)
    return oebuild



import argparse
import pytest
import os
from unittest.mock import patch, ANY


def test_setup_parser(tmp_path):
    from tuxbake.argparse import setup_parser

    assert isinstance(setup_parser(), argparse.ArgumentParser)

    """
      ( -- ) Refers to named optional arguments, i.e parser_map data can be in any order and also optional until specified as required.
    """
    build_definition = tmp_path / "oniro.json"
    build_definition.write_bytes(b'{"sources": {"kas": {}}}')
    parser_map = {
        "--build-definition": str(build_definition),
        "--runtime": "docker",
        "--image": None,
        "--src-dir": "test",
        "--build-dir-name": "build",
        "--local-manifest": None,
        "--pinned-manifest": None,
    }
    data = ["test.py"]  # adding first argument as file_name
    for key in parser_map:
        data.extend([key, parser_map[key]])
    with patch("sys.argv", data):
        data = setup_parser().parse_args()
        print(data)
        assert all(
            [
                data.build_definition == parser_map["--build-definition"],
                data.runtime == parser_map["--runtime"],
                data.image == parser_map["--image"],
                data.src_dir == os.path.abspath(parser_map["--src-dir"]),
                data.build_dir_name == parser_map["--build-dir-name"],
                data.local_manifest == parser_map["--local-manifest"],
                data.pinned_manifest == parser_map["--pinned-manifest"],
            ]
        )


def test_file_or_url(tmp_path):
    from tuxbake.argparse import file_or_url

    # case: Path to build definition file
    build_definition = tmp_path / "build-definition.json"
    build_definition.write_text("sample build definition file")
    assert file_or_url(build_definition) == build_definition

    # case: URL build definition file
    build_definition = "https:://path/to/definition_file.json"
    with patch("tuxbake.argparse.download_file") as download_file:
        download_file.return_value = tmp_path
        assert file_or_url(build_definition) == tmp_path
        assert download_file.call_count == 1
        download_file.assert_called_once_with(build_definition, ANY)

    # case: invalid path
    build_definition = "/tmp/definition.json"
    with pytest.raises(argparse.ArgumentTypeError):
        file_or_url(build_definition)

    # case: invalid URL scheme
    build_definition = "www.example.com/file/definition.json"
    with pytest.raises(argparse.ArgumentTypeError):
        file_or_url(build_definition)



""" Helper utility functions for validations of api in models.py"""

from tuxbake.exceptions import TuxbakeParsingError
import re
from typing import Any


def check_instance(value: Any, type: Any, err_msg) -> None:
    """check the type of arguments"""
    if not isinstance(value, type):
        raise TuxbakeParsingError(err_msg)


def validate_git_repo(url):
    """Validate the git_repo api argument"""
    if not url:
        raise TuxbakeParsingError("git url not provided!!")

    check_instance(
        url,
        str,
        f"Unexpected argument for 'url', expected string: '{url}'",
    )
    if "://" in url:
        url_re = re.compile(r"^(http|git|https|ssh)://[^ ]+$")
        repo_url_max_length = 1024

        if not url_re.match(url):
            raise TuxbakeParsingError(f"Unexpected argument for git url: '{url}'")

        if len(url) > repo_url_max_length:
            raise TuxbakeParsingError(
                f"url argument '{url}' too long: '{len(url)} chars'"
            )


def validate_environment(environment):
    """
    Validate environment argument which is a dictionary.
    """
    check_instance(
        environment,
        dict,
        f"Expected a dict in environment: '{environment}'",
    )
    for key, value in environment.items():
        check_instance(
            key,
            str,
            f"Unexpected argument for environment key : {key}, expected string: '{key}'",
        )
        check_instance(
            value,
            str,
            f"Unexpected argument for environment value of key : {key}, expected string: '{value}'",
        )


def validate_git_ref(git_ref):
    """Validate the git_ref api argument"""

    git_ref_re = re.compile(r"^[/\w_.-]+$")
    git_ref_max_length = 128

    check_instance(
        git_ref,
        str,
        f"Unexpected argument for git_ref, expected string: '{git_ref}'",
    )

    if not git_ref_re.match(git_ref):
        raise TuxbakeParsingError(f"Unexpected argument for git_ref: '{git_ref}'")
    if len(git_ref) > git_ref_max_length:
        raise TuxbakeParsingError(
            f"git_ref argument '{git_ref}' too long: '{len(git_ref)} chars'"
        )


def validate_git_sha(git_sha):
    """Validate the git_sha api argument"""

    git_sha_re = re.compile(r"^[0-9a-f]{40}$")

    check_instance(
        git_sha,
        str,
        f"Unexpected argument for git_sha, expected string: '{git_sha}'",
    )

    if not git_sha_re.match(git_sha):
        raise TuxbakeParsingError(
            f"Unexpected argument for git_sha: '{git_sha}'; "
            "expected 40 lowercase hexadecimal characters."
        )


def validate_git_dest(dest):
    check_instance(
        dest,
        str,
        f"Unexpected argument for dest in git_trees object, expected string: '{dest}'",
    )


def validate_container(container):
    supported_containers = [
        "ubuntu-24.04",
        "ubuntu-22.04",
        "ubuntu-20.04",
        "ubuntu-18.04",
        "ubuntu-16.04",
        "centos-7",
        "centos-8",
        "debian-bookworm",
        "debian-bullseye",
        "debian-buster",
        "debian-stretch",
        "fedora-33",
        "fedora-34",
        "opensuse-leap-15.1",
        "opensuse-leap-15.2",
    ]
    check_instance(
        container,
        str,
        f"Unexpected argument for container, expected string: '{container}'",
    )
    if container not in supported_containers:
        raise TuxbakeParsingError(
            f"Unexpected argument for container: '{container}'; "
            "container not supported or is invalid!!"
        )


def validate_local_conf(local_conf):
    check_instance(
        local_conf,
        list,
        f"Unexpected argument for local_conf, expected list: '{local_conf}'",
    )
    for conf in local_conf:
        check_instance(
            conf,
            str,
            f"Unexpected argument for local_conf data, expected string: '{conf}'",
        )


def validate_bblayers_conf(bblayers_conf):
    check_instance(
        bblayers_conf,
        list,
        f"Unexpected argument for bblayers_conf, expected list: '{bblayers_conf}'",
    )
    for bb_conf in bblayers_conf:
        check_instance(
            bb_conf,
            str,
            f"Unexpected argument for bblayers_conf data, expected string: '{bb_conf}'",
        )


def validate_artifacts(artifacts):
    check_instance(
        artifacts,
        list,
        f"Unexpected argument for artifacts, expected list: '{artifacts}'",
    )
    for artifact in artifacts:
        check_instance(
            artifact,
            str,
            f"Unexpected argument for artifacts data, expected string: '{artifact}'",
        )


def validate_extraconfigs(extraconfigs):
    check_instance(
        extraconfigs,
        list,
        f"Unexpected argument for extraconfigs, expected list: '{extraconfigs}'",
    )
    for config in extraconfigs:
        check_instance(
            config,
            str,
            f"Unexpected argument for extraconfigs data, expected string: '{config}'",
        )



import os
import json
import sys
import subprocess
import glob
from pathlib import Path
import requests
from tuxbake.exceptions import TuxbakeRunCmdError, TuxbakeParsingError
from tuxmake.logging import debug
import logging
from retrying import retry

formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)


__log_flag__ = False


def log_handler(func):
    """
    Logging decorator which toggles the log_flag to enable logging for the decorated function.
    """

    def log(*args, **kwargs):
        src_dir = args[0].src_dir
        fh = logging.FileHandler(f"{src_dir}/fetch.log")
        fh.setFormatter(formatter)
        logger.addHandler(fh)
        global __log_flag__
        __log_flag__ = True
        value = func(*args, **kwargs)
        __log_flag__ = False

        return value

    return log


@log_handler
@retry(
    wait_exponential_multiplier=500,
    wait_exponential_max=10000,
    stop_max_attempt_number=3,
    retry_on_exception=lambda e: isinstance(e, TuxbakeRunCmdError),
)
def repo_init(oebuild, src_dir, local_manifest=None, pinned_manifest=None):
    cmd = f"repo init -u {oebuild.repo.url} -b {oebuild.repo.branch} -m {oebuild.repo.manifest}".split()
    run_cmd(cmd, src_dir)
    if pinned_manifest:
        cmd = f"cp {pinned_manifest} .repo/manifests/{oebuild.repo.manifest}".split()
        run_cmd(cmd, src_dir)

    if local_manifest:
        cmd = "mkdir -p .repo/local_manifests/".split()
        run_cmd(cmd, src_dir)
        cmd = f"cp {local_manifest} .repo/local_manifests/".split()
        run_cmd(cmd, src_dir)
    cmd = "repo sync -j16".split()
    run_cmd(cmd, src_dir)
    cmd = "repo manifest -r -o pinned-manifest.xml".split()
    run_cmd(cmd, src_dir)


def git_fetch(what_to_fetch, src_dir, dir_path):
    cmd = f"git fetch origin --quiet --recurse-submodules=no {what_to_fetch}".split()
    run_cmd(cmd, os.path.join(src_dir, dir_path))


def git_get_sha(branch, url):
    cmd = f"git ls-remote --exit-code --quiet {url} {branch}".split()
    try:
        ret = run_cmd(cmd, None)
    except TuxbakeRunCmdError:
        raise TuxbakeRunCmdError(f"Unable to fetch the 'branch' or 'ref': {branch}")
    sha = ret.out.split()[0].decode("utf-8")
    return sha


@log_handler
@retry(
    wait_exponential_multiplier=500,
    wait_exponential_max=10000,
    stop_max_attempt_number=3,
    retry_on_exception=lambda e: isinstance(e, TuxbakeRunCmdError),
)
def git_init(oebuild, src_dir):
    for git_object in oebuild.git_trees:
        url = git_object.url.rstrip("/")
        branch = git_object.branch or git_object.ref
        sha = git_object.sha or git_get_sha(branch, url)
        git_object.sha = sha
        dest = git_object.dest
        if not (branch or sha):
            raise TuxbakeRunCmdError(
                f"One of branch/ref/sha should be specified for {git_object} {url}"
            )
        basename = os.path.splitext(os.path.basename(url))[0]
        # set to basename as when dest not present can be used to create source folder using basename inside src_dir
        dir_path = basename
        if dest:
            # checks ( handled ~ and ../../ )
            resolved_abs_dest = (
                os.path.abspath(os.path.join(src_dir, os.path.expanduser(dest)))
                + os.sep
            )
            if resolved_abs_dest.startswith(os.path.abspath(src_dir) + os.sep):
                dir_repo = Path(os.path.join(resolved_abs_dest, basename))
                dir_path = dir_repo
            else:
                raise TuxbakeParsingError(
                    f"Dest path provided in git_trees must be relative to src_dir: {src_dir}, curr dest: {resolved_abs_dest}"
                )
        else:
            dir_repo = Path(os.path.join(src_dir, basename))

        if not dir_repo.exists() and not dir_repo.is_dir():
            cmd = f"mkdir -p {dir_path}".split()
            run_cmd(cmd, src_dir)
            cmd = "git init --quiet .".split()
            run_cmd(cmd, os.path.join(src_dir, dir_path))
            cmd = f"git remote add origin {url}".split()
            run_cmd(cmd, os.path.join(src_dir, dir_path))

        git_fetch(sha, src_dir, dir_path)
        cmd = f"git checkout {sha}".split()
        run_cmd(cmd, os.path.join(src_dir, dir_path))

        cmd = f"git submodule update --init --recursive".split()
        run_cmd(cmd, os.path.join(src_dir, dir_path))


def get_filtered_paths(artifacts, path):
    """
    Return resolved artifacts paths which are present relative to provided path. Rest other paths are discarded.
    """
    dir_paths = []
    if artifacts and path:
        for artifact in artifacts:
            resolved_path = str(Path(f"{path}/{artifact}").resolve())
            path_files = glob.glob(resolved_path)
            if path_files and (resolved_path + os.sep).startswith(path + os.sep):
                dir_paths += path_files

    return dir_paths


def find_bitbake_env(env_file, key):
    if not os.path.exists(env_file):
        return
    with open(env_file) as env:
        for line in env.readlines():
            if line.startswith(f"{key}="):
                return line.split("=")[1].strip().replace('"', "")


def copy_artifacts(artifacts_path_list, dest_dir):
    # artifacts: list of paths
    dirs = " ".join(
        [
            path
            for path in artifacts_path_list
            if path is not None and isinstance(path, str)
        ]
    )
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    else:
        cmd = f"rm -rf {dest_dir}".split()
        run_cmd(cmd, None)
        cmd = f"mkdir -p {dest_dir}".split()
        run_cmd(cmd, None)
    cmd = f"cp -R {dirs} {dest_dir}".split()
    run_cmd(cmd, None, fail_ok=False)


def run_cmd(cmd, src_dir, env=None, fail_ok=False):
    msg = f"Running cmd: '{cmd}' in '{src_dir}'"
    debug(msg)
    process = subprocess.Popen(
        cmd, cwd=src_dir, env=env, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    process.out, process.err = process.communicate()
    handle_log([msg, process.out, process.err])
    if not fail_ok and process.returncode != 0:
        raise TuxbakeRunCmdError(f"Failed to run: {' '.join(cmd)}: {process.err}")
    return process


def handle_log(logs_list):
    global __log_flag__
    # process.err
    err = logs_list[-1]
    if err:
        print(err.decode("utf-8") if isinstance(err, bytes) else err, file=sys.stderr)

    if __log_flag__:
        for data in logs_list:
            if not data:
                continue
            elif isinstance(data, bytes):
                logger.info(data.decode("utf-8"))
            else:
                logger.info(data)
    return


def download_file(url, file_dest):
    try:
        response = requests.get(url)
        response.raise_for_status()
        json_data = response.json()
        with open(file_dest, "w") as build_definition:
            json.dump(json_data, build_definition, indent=4)
        return file_dest
    except requests.exceptions.JSONDecodeError as e:
        print(f"Invalid json file: {e}")
        sys.exit(1)
    except requests.exceptions.RequestException as e:
        print(f"Failed to download build definition file: {e}")
        sys.exit(1)



import pytest
import requests
from tuxbake.exceptions import TuxbakeRunCmdError
import os
from unittest.mock import patch, Mock


def test_run_cmd(tmp_path):
    from tuxbake.utils import run_cmd

    cmd = "ls nofile".split()
    run_cmd(cmd, src_dir=tmp_path, fail_ok=True)
    with pytest.raises(TuxbakeRunCmdError):
        run_cmd(cmd, src_dir=tmp_path, fail_ok=False)


def test_git_init(oebuild_git_object, tmpdir_factory):
    """
    oebuild_git_object is a gobal fixture defined in conftest file.
    and we are receiving it as a tuple object (oebuild_obj, src_path_1, src_path_2, git_branch_1, git_branch_2, src_dir)
    """
    from tuxbake.utils import git_init

    oebuild_object = oebuild_git_object[0]
    src_dir = oebuild_object.src_dir
    with patch("tuxbake.utils.run_cmd") as run_cmd:
        git_init(oebuild_object, src_dir)
        assert run_cmd.call_count == 12

    # case when only url is present and not branch
    for git_obj in oebuild_object.git_trees:
        # adding ref also , so as to cover ref if block
        git_obj.ref = f"refs/heads/{git_obj.branch}"
        git_obj.branch = None

    temp_src2 = tmpdir_factory.mktemp("src2")
    with patch("tuxbake.utils.run_cmd") as run_cmd:
        git_init(oebuild_object, temp_src2)
        assert run_cmd.call_count == 12

    with patch("tuxbake.utils.run_cmd", side_effect=TuxbakeRunCmdError) as run_cmd:
        with pytest.raises((TuxbakeRunCmdError, FileNotFoundError)):
            git_init(oebuild_object, "/some/wrong/folder")

    # check the retries
    for git_obj in oebuild_object.git_trees:
        git_obj.branch = None
        git_obj.sha = None
        git_obj.ref = None
    # git_init(oebuild_object, temp_src2)
    with patch("tuxbake.utils.run_cmd") as mock_run_cmd:
        with patch("tuxbake.utils.git_get_sha") as patch_get_git_sha:
            patch_get_git_sha.return_value = None
            with pytest.raises(TuxbakeRunCmdError):
                git_init(oebuild_object, temp_src2)
                assert mock_run_cmd.call_count == 3


def test_repo_init(oebuild_repo_init_object, tmpdir_factory, tmpdir):
    from tuxbake.utils import repo_init

    oebuild = oebuild_repo_init_object
    url, branch = oebuild.repo.url, oebuild.repo.branch
    temp_src = tmpdir_factory.mktemp("test_repo_init")

    # case - checking with all right parameters ( url, branch, manifest)
    with patch("tuxbake.utils.run_cmd") as run_cmd:
        repo_init(oebuild, temp_src)
        assert run_cmd.call_count == 3

    # case - checking with all right parameters with a tag.
    with patch("tuxbake.utils.run_cmd") as run_cmd:
        oebuild.repo.branch = "refs/tags/1.0.0"
        repo_init(oebuild, temp_src)
        assert run_cmd.call_count == 3

    # case - checking with wrong branch name
    oebuild.repo.branch = "some-wrong-branch"
    with patch("tuxbake.utils.run_cmd", side_effect=TuxbakeRunCmdError):
        with pytest.raises(TuxbakeRunCmdError):
            repo_init(oebuild, temp_src)

    oebuild.repo.branch = branch

    # case - checking with wrong url
    oebuild.repo.url = "https://gitlab.com/some/wrong/url/=?"
    with patch("tuxbake.utils.run_cmd", side_effect=TuxbakeRunCmdError):
        with pytest.raises(TuxbakeRunCmdError):
            repo_init(oebuild, temp_src)
    oebuild.repo.url = url

    # case - checking with local manifest file
    manifest_path = oebuild.local_manifest
    local_manifest = os.path.abspath(manifest_path)
    with patch("tuxbake.utils.run_cmd") as run_cmd:
        repo_init(oebuild, tmpdir, local_manifest)
        assert run_cmd.call_count == 5

    # case - checking with wrong manishfest file name
    oebuild.repo.manifest = "some-wrong-name.xml"
    with patch("tuxbake.utils.run_cmd", side_effect=TuxbakeRunCmdError):
        with pytest.raises(TuxbakeRunCmdError):
            repo_init(oebuild, temp_src)

    # check the retry
    # wrong manifest, will raise TuxbakeRunCmdError
    oebuild.repo.manifest = "some-wrong-name.xml"
    with patch(
        "tuxbake.utils.run_cmd", side_effect=TuxbakeRunCmdError
    ) as patch_run_cmd:
        with pytest.raises(TuxbakeRunCmdError):
            repo_init(oebuild, temp_src)
            assert patch_run_cmd.call_count == 3


def test_find_bitbake_env():
    from tuxbake.utils import find_bitbake_env

    path = os.path.abspath("tests/unit/bitbake-environment")
    assert find_bitbake_env(path, "DL_DIR")
    with pytest.raises(AssertionError):
        assert find_bitbake_env(path, "DUMMY_VAR")

    # Bitbake environment doesn't exists
    path = os.path.abspath("tests/bitbake-environment")
    assert find_bitbake_env(path, "DL_DIR") is None


def test_handle_log(capsys, oebuild_git_object):
    from tuxbake.utils import handle_log, git_init

    logs_list = ["test", b"test-check-out", b"test-check-err"]
    handle_log(logs_list)
    out, err = capsys.readouterr()
    assert "test-check-err" in err

    with capsys.disabled():
        oebuild_object = oebuild_git_object[0]
        src_dir = oebuild_object.src_dir
        git_init(oebuild_object, src_dir)

        assert os.path.exists(f"{src_dir}/fetch.log")
        with open(f"{src_dir}/fetch.log") as f:
            data = f.readline()
            assert "INFO - Running cmd:" in data


def test_copy_artifacts(tmpdir):
    from tuxbake.utils import copy_artifacts

    artifacts_path_list = ["tests/unit/conftest.py", "tests/unit/test_argparse.py"]
    copy_artifacts(artifacts_path_list, tmpdir)

    artifacts = os.listdir(tmpdir)
    assert len(artifacts) == 2
    assert "conftest.py" in artifacts
    assert "test_argparse.py" in artifacts

    artifacts_path_list = [1, 2, None]
    with pytest.raises(TuxbakeRunCmdError):
        copy_artifacts(artifacts_path_list, tmpdir)
    artifacts = os.listdir(tmpdir)
    assert len(artifacts) == 0


def test_get_filtered_paths():
    from tuxbake.utils import get_filtered_paths

    path = os.path.abspath("tests/unit/")
    artifacts = [
        "../../tuxbake/__init__.py",
        "test_utils.py",
        "../../tuxbake.spec",
        "test_build.py",
    ]
    filtered_artifacts = get_filtered_paths(artifacts, path)
    assert len(filtered_artifacts) == 2
    assert os.path.abspath("tests/unit/test_utils.py") in filtered_artifacts
    assert os.path.abspath("tests/unit/test_build.py") in filtered_artifacts

    # if path is None
    filtered_artifacts = get_filtered_paths(artifacts, None)
    assert len(filtered_artifacts) == 0

    # if artifacts is empty
    filtered_artifacts = get_filtered_paths([], path)
    assert len(filtered_artifacts) == 0


def test_download_file(tmp_path):
    from tuxbake.utils import download_file

    url = "https://www.example.com/file/definition.json"
    dest = tmp_path / "build_definition.json"
    req_mock = Mock()

    # case: successful response
    req_mock.status_code = 200
    req_mock.json.return_value = {"sources": {"repo": {}}}
    with patch("requests.get") as request:
        request.return_value = req_mock
        assert download_file(url, dest) == dest
    req_mock.json.assert_called_once()
    request.assert_called_once_with(url)
    data = open(dest).read()
    assert "repo" in data

    # case: Unsuccessful response
    req_mock.reset_mock()
    req_mock.status_code = 400
    req_mock.raise_for_status.side_effect = requests.exceptions.HTTPError(
        response=req_mock
    )
    with patch("requests.get") as request:
        request.return_value = req_mock
        with pytest.raises(SystemExit):
            download_file(url, dest)

    request.assert_called_once_with(url)
    assert req_mock.json.call_count == 0



import os
import json
import pytest
from unittest.mock import patch, MagicMock
from tuxbake.exceptions import UnsupportedMetadata
from tuxbake import __version__


def test_run_command():
    from tuxbake.metadata import run_command

    cmd = "echo hello".split()
    out = run_command(cmd, None)
    assert out == "hello"

    # case: Exception
    cmd = "catt abc.json".split()
    out = run_command(cmd, None)
    assert out is None


def test_collect(metadata_obj, tmp_path):
    from tuxbake.metadata import Metadata

    metadata_obj.src_dir = tmp_path
    with patch.object(Metadata, "generate_meta_dict"):
        with patch.object(Metadata, "free_disk_space"):
            with patch.object(Metadata, "get_git_info") as patch_git_info:
                patch_git_info.return_value = {}
                metadata_obj.collect()
    with open(f"{metadata_obj.src_dir}/metadata.json", "r") as f:
        data = json.loads(f.read())
        assert data["tuxbake"]["version"] == __version__


def test_get_runtime_info(metadata_obj):
    # update runtime
    assert metadata_obj.get_runtime_info() == {}


def test_save_metadata(metadata_obj, tmp_path):
    metadata_obj.metadata = {"a": {"key_1": "value_1", "key_2": "value_2"}}
    metadata_obj.src_dir = tmp_path
    metadata_obj.save_metadata()
    metadata = json.load(open(f"{tmp_path}/metadata.json"))
    assert metadata == metadata_obj.metadata


def test_meta_files(metadata_obj):
    available_meta_files = ["hardware", "os", "resources", "tools", "uname"]
    meta_files = metadata_obj.meta_files()
    assert sorted(meta_files) == available_meta_files


def test_generate_meta_dict(metadata_obj, tmpdir_factory):
    def prepare_ini_file(ini_data):
        # prepare a test.ini file with provided ini data
        # at temp location and return its path
        temp_dir = tmpdir_factory.mktemp("temp")
        ini_file = temp_dir / "test.ini"
        ini_file.write_text(ini_data, encoding="utf-8")
        return temp_dir

    ini_data = f"[types]\ncores=str\n\n[commands]\ncores=nproc"
    temp_dir = prepare_ini_file(ini_data)
    metadata_obj.basedir = temp_dir
    metadata_obj.files = ["test"]
    build_dir = tmpdir_factory.mktemp("build_dir")
    metadata_obj.build_dir = build_dir
    mock = MagicMock()
    mock.run_cmd.return_value = {"test": {"cores": "8"}}
    metadata_obj.runtime = mock
    with patch("shutil.copy"):
        result = metadata_obj.generate_meta_dict()
        assert result == {}

    # test exception ( without 'commands' )
    ini_data = f"[types]\ncores=str\n\n"
    temp_dir = prepare_ini_file(ini_data)
    metadata_obj.basedir = temp_dir
    metadata_obj.files = ["test"]
    with pytest.raises(UnsupportedMetadata) as msg:
        metadata_obj.generate_meta_dict()
    assert str(msg.value) == "'commands' is required in 'test.ini' file"


def test_replace_placeholders(metadata_obj):
    key = "{build_dir} placeholder check"
    value = metadata_obj.replace_placeholders(key)
    abs_path = os.path.abspath(metadata_obj.oebuild.build_dir)
    assert value == f"{abs_path} placeholder check"


def test_read_json(metadata_obj):
    # case: empty metadata json
    result = metadata_obj.read_json(None)
    assert result == {}
    # case: with 'test_key' value of type 'int'
    metadata_obj.types["test_key"] = eval("int")
    metadata_json = '{"a": {"test_key": "8"}}'
    result = metadata_obj.read_json(metadata_json)
    assert type(result["a"]["test_key"]) == int

    # case: JSONDecodeError exception
    result = metadata_obj.read_json("invalid json")
    assert result == {"invalid_metadata": "invalid json"}


def test_cast(metadata_obj):
    metadata_obj.types = {"nproc": eval("int")}
    val = metadata_obj.cast("nproc", "8")
    # int type
    assert type(val) == int
    assert val == 8


def test_free_disk_space(metadata_obj):
    metadata = metadata_obj
    with patch("shutil.disk_usage") as disk_usage:
        mock_disk_usage = MagicMock()
        mock_disk_usage.free = pow(2, 20) * 35
        disk_usage.return_value = mock_disk_usage
        assert metadata.free_disk_space() == 35


def test_get_git_info(oebuild_repo_init_object, oebuild_git_object):
    from tuxbake.metadata import Metadata

    oebuild = oebuild_repo_init_object
    # mock runtime
    runtime = MagicMock()
    runtime.source_dir = "test"
    oebuild._runtime = runtime
    metadata = Metadata(oebuild)
    # case: repo check
    with patch.object(Metadata, "meta_files") as files:
        files.return_value = []
        with patch("tuxbake.metadata.run_command") as run_cmd:
            metadata.get_git_info()
            assert run_cmd.call_count == 3

    # case: git check
    oebuild, repo1, *_ = oebuild_git_object
    # mock runtime
    runtime = MagicMock()
    runtime.source_dir = "test"
    oebuild._runtime = runtime
    metadata = Metadata(oebuild)
    # at path: repo1
    metadata.src_dir = repo1
    with patch("os.path.basename", side_effect=lambda x: ""):
        with patch("tuxbake.metadata.run_command") as run_cmd:
            metadata.get_git_info()
            assert run_cmd.call_count == 6  # 2 git repo, so 3*2 run_cmd calls



# -*- coding: utf-8 -*-

import json
import os
from dataclasses import asdict, dataclass, field, fields
from typing import Dict, List, Optional
from tuxmake.runtime import Runtime
from tuxmake.logging import set_debug, debug
from tuxbake.utils import (
    repo_init,
    git_init,
    copy_artifacts,
    get_filtered_paths,
    find_bitbake_env,
)
from pathlib import Path
from tuxbake.exceptions import TuxbakeParsingError
from tuxbake.helper import (
    check_instance,
    validate_artifacts,
    validate_bblayers_conf,
    validate_container,
    validate_environment,
    validate_extraconfigs,
    validate_git_dest,
    validate_git_ref,
    validate_git_repo,
    validate_git_sha,
    validate_local_conf,
)


class Base:
    def as_dict(self):
        return asdict(self)

    def as_json(self):
        return json.dumps(self.as_dict())

    @classmethod
    def new(cls, **kwargs):
        fields_names = [f.name for f in fields(cls)]
        i_kwargs = {}
        v_kwargs = {}
        for k in kwargs:
            if k in fields_names:
                v_kwargs[k] = kwargs[k]
            else:
                i_kwargs[k] = kwargs[k]

        return cls(**v_kwargs, extra=i_kwargs)


@dataclass
class OEBuild(Base):
    src_dir: str
    build_dir: str
    envsetup: Optional[str] = None
    targets: Optional[List[str]] = None
    after_cmds: Optional[List[str]] = None
    extraconfigs: Optional[List[str]] = None
    extraconfigs_cmd: str = ""
    distro: Optional[str] = None
    build_only: bool = False
    sync_only: bool = False
    name: Optional[str] = None
    debug: bool = False
    machine: Optional[str] = None
    container: Optional[str] = None
    environment: Dict = field(default_factory=dict)
    local_conf: Optional[List[str]] = None
    bblayers_conf: Optional[List[str]] = None
    runtime: str = "docker"
    image: Optional[str] = None
    sources: Optional[List[Dict]] = None
    sstate_dir: Optional[str] = None
    sstate_mirror: Optional[str] = None
    dl_dir: Optional[str] = None
    __logger__ = None
    repo: Optional[Dict] = None
    git_trees: Optional[List] = None
    local_manifest: Optional[str] = None
    pinned_manifest: Optional[str] = None
    # For future use
    artifacts: Optional[List[str]] = None
    artifacts_dir: Optional[str] = None
    home_dir: str = None

    @dataclass
    class Repo:
        url: str
        branch: str
        manifest: str

    @dataclass
    class Git:
        url: str
        branch: Optional[str] = None
        ref: Optional[str] = None
        sha: Optional[str] = None
        dest: Optional[str] = None

    def __post_init__(self):
        self.log_dir = self.src_dir
        if self.sources.get("repo"):
            self.repo = self.Repo(**self.sources.get("repo"))
        elif self.sources.get("git_trees"):
            self.git_trees = []
            for git_entry in self.sources.get("git_trees"):
                self.git_trees.append(self.Git(**git_entry))

    @staticmethod
    def validate(oebuild_data):
        if not oebuild_data.get("sources"):
            raise (
                TuxbakeParsingError(
                    "Please specify either git_trees or repo in sources of build definition"
                )
            )

        check_instance(
            oebuild_data["sources"],
            dict,
            f"Unexpected value of sources in build definition: {oebuild_data['sources']}, expected dictionary object with either git_trees or repo",
        )
        if oebuild_data["sources"].get("repo") and oebuild_data["sources"].get(
            "git_trees"
        ):
            raise (
                TuxbakeParsingError("repo or git_trees may be specified, but not both")
            )

        elif oebuild_data["sources"].get("repo"):
            repo = oebuild_data["sources"]["repo"]
            check_instance(
                repo,
                dict,
                f"Unexpected value of repo in sources of build definition: {repo}, expected dictionary object with url, branch and manifest",
            )
            for val in ["url", "branch", "manifest"]:
                if val not in repo.keys():
                    raise (
                        TuxbakeParsingError(
                            f"Please specify the mandatory field {val} in repo of sources"
                        )
                    )
            url, branch, manifest = repo["url"], repo["branch"], repo["manifest"]
            validate_git_repo(url)
            check_instance(
                branch,
                str,
                f"Unexpected argument for branch: {branch}, expected string: '{branch}'",
            )
            check_instance(
                manifest,
                str,
                f"Unexpected argument for manifest: {manifest}, expected string: '{manifest}'",
            )
            ext = os.path.splitext(manifest)[1]
            if manifest and ext.lower() != ".xml":
                raise (
                    TuxbakeParsingError(
                        f"unknown manifest file extension: '{ext}', must be '.xml': '{manifest}'"
                    )
                )

        elif oebuild_data["sources"].get("git_trees"):
            git_trees = oebuild_data["sources"]["git_trees"]
            check_instance(
                git_trees,
                list,
                f"Unexpected argument for git_trees: {git_trees}, expected list: '{git_trees}'",
            )

            for git_obj in git_trees:
                check_instance(
                    git_obj,
                    dict,
                    f"Unexpected values of git_trees List: {git_obj}, expected dictionary objects in the List",
                )
                if "url" not in git_obj.keys():
                    raise (
                        TuxbakeParsingError(
                            "Please specify the mandatory field url in git_trees of sources"
                        )
                    )

                url = git_obj["url"]
                branch = git_obj.get("branch", "")
                ref = git_obj.get("ref", "")
                sha = git_obj.get("sha", "")
                dest = git_obj.get("dest", "")
                # validations
                validate_git_repo(url)
                if branch:
                    check_instance(
                        branch,
                        str,
                        f"Unexpected argument for branch: {branch}, expected string: '{branch}'",
                    )
                if ref:
                    validate_git_ref(ref)
                if sha:
                    validate_git_sha(sha)
                if dest:
                    validate_git_dest(dest)
        else:
            raise (TuxbakeParsingError("repo or git_trees must be specified!!"))

        if oebuild_data.get("envsetup"):
            check_instance(
                oebuild_data["envsetup"],
                str,
                f"Unexpected argument for envsetup: {oebuild_data['envsetup']}, expected string: '{oebuild_data['envsetup']}'",
            )
        else:
            raise (
                TuxbakeParsingError("Please specify the envsetup in build definition")
            )

        if oebuild_data.get("targets"):
            check_instance(
                oebuild_data["targets"],
                list,
                f"Unexpected argument for targets: {oebuild_data['targets']}, expected list: '{oebuild_data['targets']}'",
            )
        else:
            raise (
                TuxbakeParsingError("Please specify the targets in build definition")
            )

        if oebuild_data.get("machine"):
            check_instance(
                oebuild_data["machine"],
                str,
                f"Unexpected argument for target: {oebuild_data['machine']}, expected string: '{oebuild_data['machine']}'",
            )
        else:
            raise (
                TuxbakeParsingError("Please specify the machine in build definition")
            )

        if oebuild_data.get("distro"):
            check_instance(
                oebuild_data["distro"],
                str,
                f"Unexpected argument for distro: {oebuild_data['distro']}, expected string: '{oebuild_data['distro']}'",
            )
        else:
            raise (TuxbakeParsingError("Please specify the distro in build definition"))

        if oebuild_data.get("container"):
            validate_container(oebuild_data["container"])
        else:
            raise (
                TuxbakeParsingError("Please specify the container in build definition")
            )

        if oebuild_data.get("environment"):
            validate_environment(oebuild_data["environment"])
        if oebuild_data.get("local_conf"):
            validate_local_conf(oebuild_data["local_conf"])
        if oebuild_data.get("bblayers_conf"):
            validate_bblayers_conf(oebuild_data["bblayers_conf"])
        if oebuild_data.get("extraconfigs"):
            validate_extraconfigs(oebuild_data["extraconfigs"])

        if oebuild_data.get("sstate_dir"):
            check_instance(
                oebuild_data["sstate_dir"],
                str,
                f"Unexpected argument for sstate_dir, expected string: '{oebuild_data['sstate_dir']}'",
            )
        if oebuild_data.get("sstate_mirror"):
            check_instance(
                oebuild_data["sstate_mirror"],
                str,
                f"Unexpected argument for sstate_mirror, expected string: '{oebuild_data['sstate_mirror']}'",
            )
        if oebuild_data.get("dl_dir"):
            check_instance(
                oebuild_data["dl_dir"],
                str,
                f"Unexpected argument for dl_dir, expected string: '{oebuild_data['dl_dir']}'",
            )
        if oebuild_data.get("artifacts"):
            validate_artifacts(oebuild_data["artifacts"])

    def __prepare__(self):
        os.makedirs(self.src_dir, exist_ok=True)
        os.makedirs(self.log_dir, exist_ok=True)
        debug(f"build-only flag set to: {self.build_only}")
        debug(f"sync-only flag set to: {self.sync_only}")
        if self.build_only and self.sync_only:
            print("ERROR: Both build-only and sync-only shouldn't be set.")
        if self.build_only:
            return
        if self.sources.get("repo"):
            repo_init(self, self.src_dir, self.local_manifest, self.pinned_manifest)
        else:
            git_init(self, self.src_dir)

    def prepare(self):
        set_debug(self.debug)
        self.__prepare__()
        os.environ["SKIP_OVERLAYFS"] = "true"
        self._runtime = Runtime.get(self.runtime)
        if self.sync_only:
            return
        if self.distro.startswith("openbmc"):
            self._runtime.source_dir = Path(f"{self.src_dir}/openbmc")
            self.envsetup = f"{self.envsetup} {self.machine}"
        else:
            self._runtime.source_dir = Path(self.src_dir)
        self._runtime.set_user("tuxbake")
        self._runtime.set_group("tuxbake")
        self._runtime.basename = "build"
        _container_name = f"docker.io/tuxbake/{self.container}"
        if "-local" in self.runtime:
            _container_name = f"{self.image}"
        self._runtime.set_image(f"{_container_name}")
        self._runtime.output_dir = Path(self.log_dir)
        if self.dl_dir:
            self._runtime.add_volume(self.dl_dir)
        if self.home_dir:
            self._runtime.add_volume(self.home_dir, "/home/tuxbake")
        if self.sstate_dir and not self.sstate_dir.startswith(self.src_dir):
            self._runtime.add_volume(self.sstate_dir)
        environment = self.environment
        environment["MACHINE"] = self.machine
        environment["DISTRO"] = self.distro
        self._runtime.environment = environment

        self._runtime.prepare()
        with open(
            f"{os.path.abspath(self.src_dir)}/extra_local.conf", "w"
        ) as extra_local_conf:
            if self.dl_dir:
                extra_local_conf.write(f'DL_DIR = "{self.dl_dir}"\n')
            if self.sstate_dir:
                extra_local_conf.write(f'SSTATE_DIR = "{self.sstate_dir}"\n')
            if self.sstate_mirror:
                extra_local_conf.write(f'SSTATE_MIRRORS = "{self.sstate_mirror}"\n')
                extra_local_conf.write(
                    'USER_CLASSES += "buildstats buildstats-summary"\n'
                )
            if self.local_conf:
                for line in self.local_conf:
                    extra_local_conf.write(f"{line}\n")

        if self.bblayers_conf:
            with open(
                f"{os.path.abspath(self.src_dir)}/bblayers.conf", "w"
            ) as bblayers_conf_file:
                for line in self.bblayers_conf:
                    bblayers_conf_file.write(f"{line}\n")

        if self.extraconfigs:
            copies = []
            for extra in self.extraconfigs:
                copies.append(f"cp -a {extra} conf/")
            self.extraconfigs_cmd += "&& " + " && ".join(copies)

        return

    def do_build(self):
        if self.sync_only:
            return
        # Setup bitbake environment
        cmd = [
            "bash",
            "-c",
            f"rm -f {self.build_dir}/conf/bblayers.conf && rm -f {self.build_dir}/conf/local.conf && source {self.envsetup} {self.build_dir} {self.extraconfigs_cmd} && cat ../extra_local.conf >> conf/local.conf && cat ../bblayers.conf >> conf/bblayers.conf || true && echo 'Dumping local.conf..' && cat conf/local.conf",
        ]
        if self._runtime.run_cmd(cmd, offline=False):
            self.result = "pass"
        else:
            self.result = "fail"
            return

        # Dump bitbake -e separately
        cmd = [
            "bash",
            "-c",
            f"source {self.envsetup} {self.build_dir} && bitbake -e > bitbake-environment",
        ]
        self._runtime.run_cmd(cmd, offline=False)

        for target in self.targets:
            cmd = [
                "bash",
                "-c",
                f"source {self.envsetup} {self.build_dir} && bitbake {target}",
            ]
            if self._runtime.run_cmd(cmd, offline=False):
                self.result = "pass"
            else:
                self.result = "fail"
                return

        if not self.after_cmds:
            return

        for cmd in self.after_cmds:
            cmd = [
                "bash",
                "-c",
                f"source {self.envsetup} {self.build_dir} && {cmd}",
            ]
            if self._runtime.run_cmd(cmd, offline=False):
                self.result = "pass"
            else:
                self.result = "fail"
                return

    def publish_artifacts(self):
        build_dir = f"{self._runtime.source_dir}/{self.build_dir}"
        bitbake_env_file = f"{build_dir}/bitbake-environment"
        deploy_dir = find_bitbake_env(bitbake_env_file, "DEPLOY_DIR")
        artifacts_path_list = [bitbake_env_file]
        if self.artifacts:
            build_dir_paths = get_filtered_paths(self.artifacts, build_dir)
            deploy_dir_paths = get_filtered_paths(self.artifacts, deploy_dir)
            artifacts_path_list += build_dir_paths + deploy_dir_paths
        elif deploy_dir is not None:
            # whole DEPLOY_DIR to be published
            artifacts_path_list += [f"{deploy_dir}/."]

        dest_dir = f"{self.src_dir}/{self.artifacts_dir}"
        copy_artifacts(artifacts_path_list, dest_dir)

    def do_cleanup(self):
        if self._runtime:
            self._runtime.cleanup()
